/*
 * Verarbeite eine Konfigurationsdatei
 *
 * Das meiste lesen wir einfach in ein Array. Besondere Behandlung er-
 * fahren:
 *    Alias        <url-path> <real-path>
 *    ScriptAlias  <url-path> <real-path>
 */

#ifdef _REENTRANT
#include <pthread.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pwd.h>
#include <ctype.h>
#include "HttpConfig.h"
#include "misc.h"

/*
 * Wir gehen davon aus, dass die Konfigurationsdatei nur einmal vor
 * Erzeugung anderer Threads geladen und dann nicht mehr veraendert
 * wird
 */

#ifdef _REENTRANT
pthread_mutex_t HttpConfig::mutex = PTHREAD_MUTEX_INITIALIZER;
#endif

HttpConfig::HttpConfig (void)
{
  conf = NULL;
}

HttpConfig::~HttpConfig (void)
{
  myll * c;

  while (c = conf) {
    conf = conf->next;
    free ((void *) c->attrib);
    free ((void *) c->value);
    free (c);
  }
}

void
HttpConfig::put (const char * attrib, const char * value, int type)
{
  myll * c = new myll;

  if (!c) return;
  c->attrib = strdup (attrib);
  c->value  = strdup (value);
  c->next   = conf;
  c->type   = type;
  conf      = c;
}

void
HttpConfig::AddAlias (const char * fake, const char * real)
{
  put (fake, real, ALIAS);
}

void
HttpConfig::AddScriptAlias (const char * fake, const char * real)
{
  put (fake, real, SCRIPT_ALIAS);
}

void
HttpConfig::load (const char * file)
{
  char blubber[1024];
  char * type, * ext;
  FILE * f;
  int s;

  if ((f = fopen (file, "r")) == NULL) {
    return;
  }

  while (!feof (f)) {
    if (fgets (blubber, 1024, f) == NULL)
      break;

    while (strlen(blubber)>0 && isspace(blubber[strlen(blubber)-1]))
      blubber[strlen(blubber)-1]='\0';

    if (blubber[0] == '\0' || blubber[0] == '#')
      continue;

    if ((type = strtok (blubber, " \t")) == NULL) {
      continue;
    }

    if (fpstrnicmp (type, "alias", 5)        == 0 ||
	fpstrnicmp (type, "scriptalias", 11) == 0) {
      if ((type = strtok (NULL, " \t")) != NULL) {
	type = strdup (type);
	if ((ext = strtok (NULL, " \t")) != NULL) {
	  put (type, ext, tolower(blubber[0])=='a'?ALIAS:SCRIPT_ALIAS);
	}
	free (type);
      }
    } else if (fpstrnicmp (type, "passenv", 7) == 0) {
      while ((ext = strtok (NULL, " \t")) != NULL) {
	put (type, ext);
      }
    } else {
      type = strdup (type);
      if ((ext = strtok (NULL, "\t\r\n")) != NULL) {
	put (type, ext);
      }
      free (type);
    }
  }

  fclose (f);
}

const char *
HttpConfig::operator[] (const char *str)
{
  myll * c = conf;

  while (c) {
    if (c->attrib && fpstrnicmp (str, c->attrib, strlen (str)) == 0 &&
	strlen (str) == strlen (c->attrib))
      return c->value;
    c = c->next;
  }
  return NULL;
}

char *
HttpConfig::ParseURL (const char * url, int * script, int level)
{
  char * newurl, * result = NULL;
  myll * c = conf;
  int len;

  while (c) {
    if (c->attrib == NULL || c->value == NULL ||
	(c->type != ALIAS && c->type != SCRIPT_ALIAS)) {
      c = c->next;
      continue;
    }
    if (strncmp ((char *) c->attrib, (char *) url, strlen (c->attrib)) == 0) {
      len = strlen (url) - strlen (c->attrib) + strlen (c->value) + 1;
      if ((newurl = (char *) malloc (len)) == NULL) {
	return NULL;
      }

      sprintf (newurl, "%s%s", c->value, url+strlen(c->attrib));

      if (c->type == SCRIPT_ALIAS) {
	if (script) *script = 1;
	return newurl;
      }
      if (script) *script = 0;
      if ((result = (char *) ParseURL (newurl, script, level+1)) != newurl) {
	free (newurl);
      }
      return result;
    }
    c = c->next;
  }

  if (script) *script = 0;

  result = (char *) url;

  /*
   * Home-Verzeichnis aufloesen
   */

  if (result[0] == '/' && result[1] == '~' && result[2] != '\0') {
    struct passwd *pw;
    const char * ud;
    char user[32];
    int nlen, hpw;

    if (strchr (result+1, '/')) {
      nlen = strchr (result+1, '/') - result - 2;
    } else {
      nlen = strlen (result+2);
    }
    if (nlen > 31) return NULL;
    memcpy (user, result+2, nlen);
    user[nlen] = '\0';

    ud = (*this)["UserDir"];

    /*
     * getpwnam ist nicht reentrant
     */

#ifdef _REENTRANT
    pthread_mutex_lock (&mutex);
#endif
    pw = getpwnam (user);
    
    if (pw) {
      len = strlen(result) + strlen (pw->pw_dir) + 1;
      if (ud != NULL) {
	len += strlen (ud);
      }
      if ((newurl = (char *) malloc (len)) == NULL) {
	if (result != url) free (result);
#ifdef _REENTRANT
	pthread_mutex_unlock (&mutex);
#endif
	return NULL;
      }
      sprintf (newurl, "%s%s%s%s",
	       pw->pw_dir, ud ? "/" : "", ud ? ud : "",
	       result+nlen+2);
      
      if (result != url) free (result);
      result = newurl;
    }
    else {
      printf ("getpwnam `%s' fehlgeschlagen\n", user);
    }
#ifdef _REENTRANT
    pthread_mutex_unlock (&mutex);
#endif
  }

  /*
   * Wenn noch nichts aufgeloest wurde, nutze die DocumentRoot
   */

  if (url == result && level == 0 && (*this)["DocumentRoot"] != NULL) {
    len = strlen (url) + strlen ((*this)["DocumentRoot"]) + 1;
    if ((result = (char *) malloc (len)) == NULL) {
      return NULL;
    }
    sprintf (result, "%s%s", (*this)["DocumentRoot"],
	     url[0] ? url : url+1);
  } else if (url == result && level == 0 && url[0] == '/' && url[1]) {
    len = strlen (url);
    if ((result = (char *) malloc (len)) == NULL) {
      return NULL;
    }
    strcpy (result, url+1);
  }

  /*
   * Wenn auf ein Verzeichnis zugegriffen wird, haenge DirectoryIndex
   * hintendran
   */

  if (result[0] && result[strlen(result)-1] == '/' &&
      (*this)["DirectoryIndex"] != NULL) {
    len = strlen (result) + strlen ((*this)["DirectoryIndex"]) + 1;

    if ((newurl = (char *) malloc (len)) == NULL) {
      free (result);
      return NULL;
    }
    sprintf (newurl, "%s%s", result, (*this)["DirectoryIndex"]);

    if (result != url) free (result);
    result = newurl;
  }

  return result;
}

int
HttpConfig::Uploadable (const char * url)
{
  myll * c = conf;

  while (c) {
    if (fpstrnicmp ((char *) c->attrib, "Uploadable", 10) == 0) {
      if (strncmp ((char *) c->value, (char *) url, strlen (c->value)) == 0) {
	return 1;
      }
    }
    c = c->next;
  }
  return 0;
}

char **
HttpConfig::PassEnviron (char **env)
{
  myll * c = conf;

  while (c) {
    if (fpstrnicmp ((char *) c->attrib, "PassEnv", 7) == 0) {
      if (c->value && getenv (c->value)) {
	fpaddenv (&env, c->value, getenv (c->value));
      }
    }
    c = c->next;
  }

  return env;
}
