/*
 * Bediene einen HTTP Request
 */

#ifdef _REENTRANT
#include <pthread.h>
#endif

#include <sys/wait.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <time.h>
#include <signal.h>
#include <errno.h>
#include <socket_class.h>
#include "HttpHeader.h"
#include "HttpConfig.h"
#include "MimeTypes.h"
#include "fpSocket.h"
#include "misc.h"
#include "HttpServer.h"
#include "HttpRequest.h"


int HttpRequest::counter = 0;

/*
 * ----------------------------------------------------------------------
 * HttpRequest
 * ----------------------------------------------------------------------
 */

HttpRequest::HttpRequest (fpSocket * s, HttpServer * m)
  : sock(s), master(m)
{
  script     = 0;
  httpvers   = 0;
  origURL    = NULL;
  decodedURL = NULL;
  targetURL  = NULL;
  instance   = counter++;
}

HttpRequest::~HttpRequest (void)
{
  if (targetURL != decodedURL) {
    free (targetURL);
  }
  free (decodedURL);
  free (origURL);
}

void
HttpRequest::PrintLog (const char * fmt, ...)
{
  va_list args;
  va_start(args, fmt);

  if (master->log) {
#ifdef _REENTRANT
    pthread_mutex_lock (&master->mutex);
#endif
    fprintf  (master->log, "HttpRequest: %4d: ", instance);
    vfprintf (master->log, fmt, args);
    fputc    ('\n', master->log);
    fflush   (master->log);
#ifdef _REENTRANT
    pthread_mutex_unlock (&master->mutex);
#endif
  }

  va_end (args);
}

void
HttpRequest::AccessLog (int code, long total)
{
  if (master->transfer_log) {
    char blubber[32], hostname[128];
    struct tm ltime;
    time_t now;

    now   = time (NULL);
    fp_localtime_r (&now, &ltime);

    strftime (blubber, 32, "%d/%b/%Y:%H:%M:%S", &ltime);
    blubber[31] = '\0';

#ifdef _REENTRANT
    pthread_mutex_lock (&master->mutex);
#endif

    fprintf (master->transfer_log,
	     "%s - - [%s +0000] \"%s %s\" %03d %ld\n",
	     sock->gethostname_r (hostname, 128)?
	     sock->gethostname_r (hostname, 128):
	     sock->getinetaddr_r (hostname, 128),
	     blubber, mname, origURL, code, total);
    fflush (master->transfer_log);

#ifdef _REENTRANT
    pthread_mutex_unlock (&master->mutex);
#endif
  }
}

long
HttpRequest::DumpFile (FILE * file)
{
  long count, total=0;
  char buf[256];

  while (!feof (file) && !!*sock) {
    if ((count = fread (buf, 1, 256, file)) == 0) {
      break;
    }
    sock->write (buf, count);
    total += count;
  }
  return total;
}

void
HttpRequest::DumpError (const char * msg)
{
  char temp[100];

  if (!*sock) return;
  if (httpvers >= 10) {
    sock->printf ("HTTP/1.1 %s\r\n", msg);
    sock->printf ("Content-Type: text/html\r\n");
    sock->printf ("Server: %s\r\n", master->ServerName);
    sock->printf ("Date: %s\r\n", HttpMakeDate (temp, 100));
    sock->printf ("\r\n");
  }
  sock->printf ("<html><head>\r\n\
<title>Error retrieving document</title>\r\n\
</head>\r\n\
<body>\r\n\
<h1>Error retrieving document</h1>\r\n\
%s\r\n\
<p>\r\n\
Sorry about that, I'm just doing my job.\r\n\
<p>\r\n\
Your most humble fpServ agent.\r\n\
</body>\r\n\
</html>\r\n\
", msg);
}

/*
 * Bediene ein Skript, GET oder HEAD
 */

void
HttpRequest::ServeScript (void)
{
  char *path, *ptr, *epath, *qs, old='\0';
  char *scriptname, *progname, temp[256];
  int in[2], out[2], count, total;
  struct stat buf;
  pid_t pid;

  if ((path = strdup (targetURL)) == NULL) {
    DumpError ("500 Internal server error");
    return;
  }

  ptr = path + 1;

  while (*ptr) {
    while (*ptr && *ptr!='/') ptr++;
    old  = *ptr;
    *ptr = '\0';

    if (stat (path, &buf) == 0) {
      if ((buf.st_mode & S_IXUSR) && !S_ISDIR(buf.st_mode)) {
	break;
      }
    }

    *ptr = old;

    if (!*ptr) {
      PrintLog  ("could not find script for `%s'", targetURL);
      DumpError ("404 File not found");
      AccessLog (404, 0);
      free (path);
      return;
    }

    ptr++;
  }

  PrintLog ("script is `%s'", path);

  /*
   * Extrapfad
   */

  if (old) {
    if ((epath = (char *) malloc (strlen (ptr+1) + 2)) == NULL) {
      DumpError ("500 Internal server error");
      free (path);
      return;
    }

    epath[0] = '/';
    strcpy (epath+1, ptr+1);

    if (strchr (epath, ' ')) *(strchr (epath, ' ')) = '\0';
    if (strchr (epath, '?')) *(strchr (epath, '?')) = '\0';
    PrintLog ("extra path is `%s'", epath);
  } else {
    epath = NULL;
  }

  /*
   * Query String
   */

  if (strchr (targetURL, '?')) {
    if ((qs = strdup (strchr (targetURL, '?') + 1)) == NULL) {
      DumpError ("500 Internal server error");
      free (epath);
      free (path);
      return;
    }
    PrintLog ("query string is `%s'", qs);
  } else {
    qs = NULL;
  }

  /*
   * Scriptname
   */

  if ((progname = strrchr (path, '/')) != NULL) {
    if (strstr (decodedURL, progname+1) != NULL) {
      if ((scriptname = strdup (decodedURL)) == NULL) {
	DumpError ("500 Internal server error");
	free (epath);
	free (path);
	free (qs);
      }
      *(strstr(scriptname,progname+1)+strlen(progname+1)) = '\0';
      PrintLog ("scriptname is `%s'", scriptname);
    } else {
      scriptname = NULL;
    }
  } else {
    scriptname = NULL;
  }

  /*
   * Zwei Pfeifen fuer die Kommunikation
   */

  if (pipe(in) != 0) {
    PrintLog  ("could not create pipe");
    DumpError ("500 Internal server error");
    free (scriptname);
    free (epath);
    free (path);
    free (qs);
    return;
  }

  if (pipe(out) != 0) {
    PrintLog  ("could not create pipe");
    DumpError ("500 Internal server error");
    close (in[0]);
    close (in[1]);
    free (scriptname);
    free (epath);
    free (path);
    free (qs);
    return;
  }

  PrintLog ("forking ...");

  if ((pid = fork()) == (pid_t) -1) {
    PrintLog  ("could not fork");
    DumpError ("500 Internal server error");
    close (in[0]); close (out[0]);
    close (in[1]); close (out[1]);
    free (scriptname);
    free (epath);
    free (path);
    free (qs);
    return;
  }

  if (pid == 0) {
    /*
     * Kindprozess
     */
    close (in[1]);
    close (out[0]);
    //    close (0);
    //    close (1);
    dup2  (in[0], 0);
    dup2  (out[1], 1);

    char **env = NULL, **ei;
    char blubb[256];

    fpaddenv (&env, "PATH", "/bin:/usr/bin");

    env = master->conf.PassEnviron (env);

    fpaddenv (&env, "SERVER_SOFTWARE",   master->ServerName);
    fpaddenv (&env, "GATEWAY_INTERFACE", "CGI/1.1");

    sprintf  (blubb, "HTTP/%d.%d", httpvers/10, httpvers%10);
    fpaddenv (&env, "SERVER_PROTOCOL",   blubb);

    sprintf  (blubb, "%d", (int) master->portno);
    fpaddenv (&env, "SERVER_PORT",       blubb);

    /*
     * We must not use the reentrant interface here, because the
     * mutexes may be inherited in a locked state
     */

    fpaddenv (&env, "REMOTE_ADDR", sock->getpeeraddr()->getinetaddr ());

    if (sock->getpeeraddr()->gethostname ()) {
      fpaddenv (&env, "REMOTE_HOST", sock->getpeeraddr()->gethostname ());
    }

    fpaddenv (&env, "PATH_INFO",       epath?epath:"");
    fpaddenv (&env, "QUERY_STRING",    qs?qs:"");
    fpaddenv (&env, "REQUEST_METHOD",  (method==HTTP_GET)?"GET":"POST");
    fpaddenv (&env, "SCRIPT_NAME",     scriptname?scriptname:"");

    if (method == HTTP_POST) {
      fpaddenv (&env, "CONTENT_LENGTH",
		headers["Content-Length"]?headers["Content-Length"]:"0");
    }
    if (headers["Content-Type"]) {
      fpaddenv (&env, "CONTENT_TYPE", headers["Content-Type"]);
    }
    if (headers["User-Agent"]) {
      fpaddenv (&env, "HTTP_USER_AGENT", headers["User-Agent"]);
    }
    if (headers["Cookie"]) {
      fpaddenv (&env, "HTTP_COOKIE", headers["Cookie"]);
    }

    /*
     * Fire away
     */

    execle (path, path, NULL, env);
    exit (1);
  }

  /*
   * Vaterprozess
   */

  PrintLog ("father is alive and well");

  close (in[0]);
  close (out[1]);

  if (method == HTTP_POST) {
    if (headers["Content-Length"])
      total = atoi (headers["Content-Length"]);
    else
      total = 0;

    while (total && !!*sock) {
      count = sock->read (temp, (total>256)?256:total);
      if (write (in[1], temp, count) == -1) break;
      total -= count;
    }

    if (!*sock || total) {
      kill  (pid, SIGTERM);
      close (out[0]);
      PrintLog  ("error dumping post data");
      DumpError ("500 Internal server error");
      free (scriptname);
      free (epath);
      free (path);
      free (qs);
      return;
    }

    close (in[1]);
  }

  /*
   * Sieht ja soweit ganz gut aus
   */

  PrintLog ("serving ...");

  if (httpvers >= 10) {
    sock->printf ("HTTP/1.1 200 OK\r\n");
    sock->printf ("Server: %s\r\n", master->ServerName);
    sock->printf ("Date: %s\r\n", HttpMakeDate (temp, 256));
  }

  total = 0;

  while ((count = read (out[0], temp, 256))) {
    sock->write (temp, count);
    total += count;
  }

  close (out[0]);
  wait (NULL);

  /*
   * Phew
   */

  PrintLog  ("done (%ld bytes served).", total);
  AccessLog (200, count);

  free (scriptname);
  free (epath);
  free (path);
  free (qs);
}

/*
 * Bediene einen Get oder Head Request
 */

void
HttpRequest::ServeGet (void)
{
  const char *ext, *type;
  struct stat buf;
  struct tm ttm;
  char temp[100];
  long count;
  FILE *f;

  if (script) {
    ServeScript();
    return;
  }

  /*
   * Dump a file
   */

  if (stat (targetURL, &buf) != 0) {
    PrintLog  ("could not stat file `%s': %s",
	       targetURL, strerror (errno));
    DumpError ("404 File not found");
    AccessLog (404, 0);
    return;
  }

  if (S_ISDIR(buf.st_mode)) {
    PrintLog  ("`%s' is a directory, redirecting to `%s/'",
	       targetURL, origURL);
    if (httpvers >= 10) {
      sock->printf ("HTTP/1.1 301 Moved Permanently\r\n");
      sock->printf ("Server: %s\r\n", master->ServerName);
      sock->printf ("Location: %s/\r\n", origURL);
      sock->printf ("Date: %s\r\n", HttpMakeDate (temp, 100));
      sock->printf ("\r\n");
    }
    sock->printf ("<html><head>\r\n\
<title>Document has moved</title>\r\n\
</head>\r\n\
<body>\r\n\
<h1>Document has moved</h1>\r\n\
Click for the <a href=\"%s/\">New Location</a>\r\n\
<p>\r\n\
Your most humble fpServ agent.\r\n\
</body>\r\n\
</html>\r\n\
", origURL);
    AccessLog (301, 0);
    return;
  }

  if ((f = fopen (targetURL, "r")) == NULL) {
    PrintLog  ("could not open file `%s': %s",
	       targetURL, strerror (errno));
    DumpError ("404 File not found");
    AccessLog (404, 0);
    return;
  }

  if (headers["If-Modified-Since"]) {
    time_t ims = HttpScanDate (headers["If-Modified-Since"]);

    if (buf.st_mtime <= ims) {
      if (httpvers >= 10) {
	sock->printf ("HTTP/1.1 304 Not Modified\r\n");
	sock->printf ("Server: %s\r\n", master->ServerName);
	sock->printf ("Date: %s\r\n", HttpMakeDate (temp, 100));
	sock->printf ("\r\n");
      }
      PrintLog  ("Document not modified since %s",
		 headers["If-Modified-Since"]);
      AccessLog (304, 0);
      return;
    }
  }
  
  if (httpvers >= 10) {
    sock->printf ("HTTP/1.1 200 Ok\r\n");
    sock->printf ("Server: %s\r\n", master->ServerName);
    sock->printf ("Date: %s\r\n", HttpMakeDate (temp, 100));
    
    ext = targetURL + strlen (targetURL) - 1;
    
    while (ext>targetURL && *ext != '.')
      ext--;
    
    if (master==NULL || (type = master->mime[ext+1]) == NULL) {
      sock->printf ("Content-Type: application/octet-stream\r\n");
    } else {
      PrintLog     ("content type is `%s'", type);
      sock->printf ("Content-Type: %s\r\n", type);
    }

    sock->printf ("Content-Length: %d\r\n", (int) buf.st_size);
    sock->printf ("Last-Modified: %s\r\n", HttpMakeDate (temp, 100, buf.st_mtime));
    sock->printf ("\r\n");

    if (method == HTTP_HEAD) {
      AccessLog (200, 0);
      return;
    }
  }
  
  PrintLog  ("serving ...");
  count = DumpFile (f);
  PrintLog  ("done (%ld bytes served).", count);
  AccessLog (200, count);
  fclose (f);
  return;
}

/*
 * Bediene einen Put Request
 */

void
HttpRequest::ServePut (void)
{
  char temp[1024];
  long count, total;
  int existed;
  struct stat buf;
  FILE * out;

  if (!headers["Content-Length"]) {
    PrintLog  ("no Content-Length header for put request");
    DumpError ("411 Length Required");
    AccessLog (411, 0);
    return;
  }

  if (!master->conf.Uploadable (targetURL)) {
    PrintLog  ("directory `%s' not uploadable", targetURL);
    DumpError ("405 Method not Allowed");
    AccessLog (405, 0);
    return;
  }

  existed = ! stat (targetURL, &buf);

  if ((out = fopen (targetURL, "w")) == NULL) {
    PrintLog  ("could not open file `%s' for writing: %s",
	       targetURL, strerror (errno));
    DumpError ("403 Forbidden");
    AccessLog (403, 0);
    return;
  }

  total = atoi (headers["Content-Length"]);

  PrintLog ("receiving %d bytes for put", total);

  while (total) {
    if ((count = sock->read (temp, 1024)) == -1)
      break;
    fwrite (temp, 1, count, out);
    total -= count;
  }

  fclose (out);

  if (total) {
    unlink    (targetURL);
    PrintLog  ("oops: %d bytes of %s left to read",
	       total, headers["Content-Length"]);
    DumpError ("400 premature eof");
    AccessLog (400, 0);
    return;
  }

  if (existed) {
    sock->printf ("HTTP/1.1 200 Ok\r\n");
  } else {
    sock->printf ("HTTP/1.1 201 Created\r\n");
  }

  sock->printf ("Server: %s\r\n", master->ServerName);
  sock->printf ("Date: %s\r\n", HttpMakeDate (temp, 100));
  sock->printf ("Content-Type: text/plain\r\n");
  sock->printf ("\r\n");
  sock->printf ("Document uploaded.\r\n");

  if (existed) {
    AccessLog (200, 0);
  } else {
    AccessLog (201, 0);
  }
}

/*
 * Bediene einen Delete Request
 */

void
HttpRequest::ServeDelete (void)
{
  struct stat buf;
  char temp[100];

  if (!master->conf.Uploadable (targetURL)) {
    PrintLog  ("directory for `%s' not uploadable", targetURL);
    DumpError ("405 Method Not Allowed");
    AccessLog (403, 0);
    return;
  }

  if (stat (targetURL, &buf) != 0) {
    PrintLog  ("to-be-deleted file `%s' does not exist", targetURL);
    DumpError ("404 Not Found");
    AccessLog (404, 0);
    return;
  }

  if (unlink (targetURL) != 0) {
    PrintLog  ("cannot unlink `%s': %s", targetURL, strerror (errno));
    DumpError ("403 Forbidden");
    AccessLog (403, 0);
    return;
  }

  sock->printf ("HTTP/1.1 200 Ok\r\n");
  sock->printf ("Server: %s\r\n", master->ServerName);
  sock->printf ("Date: %s\r\n", HttpMakeDate (temp, 100));
  sock->printf ("Content-Type: text/plain\r\n");
  sock->printf ("\r\n");
  sock->printf ("Resource deleted.\r\n");
  AccessLog (200, 0);
}

/*
 * Bediene einen Post Request
 */

void
HttpRequest::ServePost (void)
{
  ServeScript ();
}

/*
 * Bediene einen einzelnen HTTP Request
 */

int
HttpRequest::serve (void)
{
  char intro[4096], *ptr, *p2;
  char hostname[128];
  int v1, v2;

  PrintLog ("new connection, peer %s:%d",
	    sock->gethostname_r (hostname, 128),
	    (int) sock->getpeeraddr()->getport());

  if (sock->fgets (intro, 4096) == NULL) {
    PrintLog  ("error: could not read request");
    DumpError ("400 Bad Request");
    return -1;
  }

  intro[4095] = '\0';

  if (HttpServer::logrequests) {
    PrintLog ("input `%s'", intro);
  }

  if (strlen (intro) == 4095) {
    PrintLog  ("oops, input too long");
    DumpError ("414 Request-URI Too Large");
    return -1;
  }

  if ((ptr = strchr (intro, ' ')) == NULL) {
    PrintLog  ("error: illegal request: `%s'", intro);
    DumpError ("400 Bad Request");
    return -1;
  } else {
    *ptr = '\0';
  }

  if (fpstrnicmp (intro, "get", 3) == 0)
    method = HTTP_GET;
  else if (fpstrnicmp (intro, "head", 4) == 0)
    method = HTTP_HEAD;
  else if (fpstrnicmp (intro, "post", 4) == 0)
    method = HTTP_POST;
  else if (fpstrnicmp (intro, "put", 3) == 0)
    method = HTTP_PUT;
  else if (fpstrnicmp (intro, "delete", 6) == 0)
    method = HTTP_DELETE;
  else {
    PrintLog  ("error: illegal method: `%s'", intro);
    DumpError ("501 Not Implemented");
    return -1;
  }

  strcpy (mname, intro);

  /*
   * URL
   */

  if ((p2 = strchr (++ptr, ' ')) != NULL) {
    *p2 = '\0';
  }

  if ((origURL = strdup (ptr)) == NULL) {
    PrintLog  ("oops: could not dup `%s'", ptr);
    DumpError ("500 Internal Server Error");
    return -1;
  }

  PrintLog ("incoming %s request for `%s'",
	    mname, origURL);

  decodedURL = HttpUnEncode (origURL);

  PrintLog ("request decoded as `%s'", decodedURL);

  /*
   * Rewrite URL
   */
  if (master) {
    if ((targetURL = master->conf.ParseURL (decodedURL, &script)) == NULL) {
      PrintLog  ("oops: rewriting URL failed");
      DumpError ("500 Internal Server Error");
      return -1;
    }
  } else {
    script    = 0;
    targetURL = decodedURL;
  }

  PrintLog ("URL rewritten as `%s'", targetURL);

  if (script) {
    PrintLog ("wow, this is a script");
  }

  /*
   * Determine HTTP version
   */

  if (p2) {
    p2++;
    if ((sscanf (p2, "HTTP/%d.%d", &v1, &v2)) == 2) {
      httpvers = v1*10 + v2;
    } else {
      httpvers = 0;
    }
  } else {
    httpvers = 9;
  }
  
  PrintLog ("client speaks HTTP version %d.%d", httpvers/10, httpvers%10);

  if (httpvers >= 10) {
    /*
     * HTTP/1.x, Read Headers
     */
    if (sock->fgets (intro, 4096) != NULL) {
      if (HttpServer::logrequests) {
	PrintLog ("input `%s'", intro);
      }
      while (intro[0] && !!*sock) {
	headers.put (intro);
	if (sock->fgets (intro, 4096) == NULL)
	  break;
	if (HttpServer::logrequests) {
	  PrintLog ("input `%s'", intro);
	}
      }
    }
  }

  switch (method) {
  case HTTP_GET:
    ServeGet ();
    break;

  case HTTP_POST:
    ServePost ();
    break;

  case HTTP_HEAD:
    ServeGet ();
    break;

  case HTTP_PUT:
    ServePut ();
    break;

  case HTTP_DELETE:
    ServeDelete ();
    break;
  }

  return 0;
}
