/*
 * Test an HTTP Server
 */

#ifndef _REENTRANT
#define _REENTRANT
#endif

#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

int             theend;
int             opensocks;
int             maxsock;
int             failed;
fd_set          sockset;
pthread_mutex_t mutex;
pthread_cond_t  cond;

int totalmax, hits;
unsigned long transferred;

int
closethesock (int desc)
{
  int index = desc;

  pthread_mutex_lock (&mutex);
  FD_CLR(index, &sockset);
  if (index == maxsock) {
    for (--index; index>=0; index--) {
      if (FD_ISSET(index, &sockset)) {
	maxsock = index;
	break;
      }
    }
  }
  opensocks--;
  pthread_cond_broadcast (&cond);
  pthread_mutex_unlock (&mutex);

  return 0;
}

void *
consumer (void * dummy)
{
  char buffer[1024];
  fd_set myset;
  int maxdesc;
  int res, index;
  int count;

#ifdef DEBUG
  unsigned long tcount;
  int rdescs;
#endif

#ifdef DEBUG
  fprintf (stdout, "consumer hello\n");
#endif

  while (!theend || opensocks) {
    pthread_mutex_lock (&mutex);
    while (!theend && !opensocks) {
      pthread_cond_wait (&cond, &mutex);
    }
    if (theend && !opensocks) {
      pthread_mutex_unlock (&mutex);
      break;
    }

    maxdesc = maxsock;
    memcpy (&myset, &sockset, sizeof (fd_set));

    pthread_mutex_unlock (&mutex);

#ifdef DEBUG
    fprintf (stdout, "select test: ");
    for (index=0; index<maxdesc; index++) {
      fprintf (stdout, "%d ", FD_ISSET(index, &myset));
    }
    fprintf (stdout, "\n");
#endif

    if ((res = select (maxdesc, &myset, NULL, NULL, NULL)) <= 0) {
      continue;
    }

#ifdef DEBUG
    fprintf (stdout, "select res:  ");
    for (index=0; index<maxdesc; index++) {
      fprintf (stdout, "%d ", FD_ISSET(index, &myset));
    }
    fprintf (stdout, "\n");
    rdescs = 0;
    tcount++;
#endif

    for (index=0; index<maxdesc; index++) {
      if (FD_ISSET(index, &myset)) {
	if ((count = read (index, buffer, 1024)) == 0) {
	  closethesock (index);
	}
	if (count > 0) {
	  transferred += count;
#ifdef DEBUG
	  rdescs++;
	  tcount+=count;
#endif
	}
      }
    }

#ifdef DEBUG
    fprintf (stdout, "read %ld bytes from %d descs\n", tcount, rdescs);
#endif
  }

  return NULL;
}

int
newhit (const char *address, unsigned short portno, const char *http)
{
  struct sockaddr_in addr;
  struct hostent * nodep;
  int desc;

  if ((desc = socket (AF_INET, SOCK_STREAM, 0)) == -1) {
    return -1;
  }

  memset (&addr, 0, sizeof (struct sockaddr_in));

  addr.sin_family = AF_INET;
  addr.sin_port   = htons (portno);

  if ((addr.sin_addr.s_addr = inet_addr (address)) == INADDR_NONE) {
    if ((nodep = gethostbyname (address)) == NULL) {
      close (desc);
      return -1;
    }
    memcpy (&addr.sin_addr.s_addr, nodep->h_addr, nodep->h_length);
  }

  if (connect (desc, (struct sockaddr *) &addr, sizeof (struct sockaddr_in)) == -1) {
    close (desc);
    return -1;
  }

  if (write (desc, http, strlen (http)) != strlen (http)) {
    close (desc);
    return -1;
  }

  pthread_mutex_lock (&mutex);

  FD_SET(desc, &sockset);
  if (desc+1 > maxsock)
    maxsock = desc+1;
  if (++opensocks > totalmax)
    totalmax = opensocks;
  hits++;
  pthread_cond_broadcast (&cond);
  pthread_mutex_unlock (&mutex);

#ifdef DEBUG
  fprintf (stderr, "new hit %d, desc %d maxsock %d open %d\n",
	   hits, desc, maxsock, opensocks);
#endif
  return 0;
}

int
producer (const char *server, unsigned short portno, const char *url,
	  int count, int parallel)
{
  char buffer[1024];
  int running;

  sprintf (buffer, "GET %s HTTP/1.0\r\n\r\n", url);

  while (42) {
    pthread_mutex_lock (&mutex);

    if (hits == count) {
      theend = 1;
      pthread_cond_broadcast (&cond);
      pthread_mutex_unlock (&mutex);
      break;
    }
    running = opensocks;

    pthread_mutex_unlock (&mutex);

    if (running < parallel) {
      if (newhit (server, portno, buffer) == -1) {
	if (errno == EMFILE) {
	  pthread_mutex_lock (&mutex);
	  while (opensocks && running == opensocks) {
	    pthread_cond_wait (&cond, &mutex);
	  }
	  pthread_mutex_unlock (&mutex);
	}
	else if (errno == ETIMEDOUT || errno == ECONNREFUSED) {
	  pthread_mutex_lock (&mutex);
	  failed++;
	  hits++;
	  while (opensocks && running == opensocks) {
	    pthread_cond_wait (&cond, &mutex);
	  }
	  pthread_mutex_unlock (&mutex);
	}
	else {
	  fprintf (stderr, "oops: newhit errno %d\n", errno);
	}
      }
    }
    
    if (running == parallel) {
      pthread_mutex_lock (&mutex);
      while (opensocks && opensocks >= parallel) {
	pthread_cond_wait (&cond, &mutex);
      }
      pthread_mutex_unlock (&mutex);
    }    
  }
  return 0;
}

int
main (int argc, char *argv[])
{
  char hostname[1024], url[1024], port[1024], *ptr;
  unsigned short portno;
  int index, howmuch, howmax;
  pthread_t th;
  struct timeval start, ende;
  double seconds, hps, kbps;

  if (argc < 3 || argc > 4) {
    fprintf (stderr, "usage: %s <url> <hits> [<parallel>]\n", argv[0]);
    exit (1);
  }

  if (strncmp (argv[1], "http://", 7) == 0) {
    ptr = argv[1] + 7;
  }
  else {
    ptr = argv[1];
  }

  ptr = argv[1] + 7;
  for (index=0; ptr[index] && ptr[index]!=':' && ptr[index]!='/'; index++)
    hostname[index] = ptr[index];
  hostname[index] = '\0';
  ptr += index;

  if (*ptr == ':') {
    ptr++;
    for (index=0; ptr[index] && ptr[index]!='/'; index++)
      port[index] = ptr[index];
    port[index] = '\0';
    ptr += index;
  }
  else {
    strcpy (port, "80");
  }

  if ((portno = (unsigned short) strtol (port, NULL, 0)) == 0) {
    fprintf (stderr, "%s; illegal port number\n", port);
    exit (1);
  }

  if (*ptr == '/') {
    strcpy (url, ptr);
  }
  else {
    strcpy (url, "/");
  }

  if ((howmuch = strtol (argv[2], NULL, 0)) <= 0) {
    fprintf (stderr, "%s: not a valid number of hits\n", argv[2]);
    exit (1);
  }

  if (argc == 4) {
    if ((howmax = strtol (argv[3], NULL, 0)) <= 0) {
      fprintf (stderr, "%s: not a valid number of parallel hits\n",
	       argv[3]);
      exit (1);
    }
  }
  else {
    howmax = 1;
  }

  fprintf (stdout, "bombing http://%s:%d%s with %d hits (%d parallel)\n",
	   hostname, (int) portno, url, howmuch, howmax);

  /*
   * Initialisierung
   */

  theend    = 0;
  opensocks = 0;
  maxsock   = 0;
  failed    = 0;
  totalmax  = 0;
  hits      = 0;
  transferred = 0;

  FD_ZERO (&sockset);
  pthread_cond_init (&cond, NULL);
  pthread_mutex_init (&mutex, NULL);

  /*
   * consumer starten
   */

  pthread_create (&th, NULL, consumer, NULL);

  /*
   * Und los gehts
   */

  gettimeofday (&start, NULL);
  producer (hostname, portno, url, howmuch, howmax);
  pthread_join (th, NULL);
  gettimeofday (&ende, NULL);
  
  /*
   * Auswerten
   */

  seconds = (ende.tv_sec-start.tv_sec) +
    ((double) (ende.tv_usec-start.tv_usec) / 1000000.0);

  hps  = hits / seconds;
  kbps = transferred / seconds / 1024.0;

  fprintf (stdout, "\n");
  fprintf (stdout, "     Total Hits: %10d\n",   hits);
  fprintf (stdout, "  Max. parallel: %10d\n",   totalmax);
  fprintf (stdout, " Bytes received: %10ld\n",  transferred);
  fprintf (stdout, "    Time passed: %10.3f seconds\n", seconds);
  fprintf (stdout, "\n");
  fprintf (stdout, "         hits/s: %10.3f\n", hps);
  fprintf (stdout, "       hits/min: %10.3f\n", hps*60);
  fprintf (stdout, "         hits/h: %10.3f\n", hps*3600);
  fprintf (stdout, "\n");
  fprintf (stdout, "           kB/s: %10.3f\n", kbps);
  fprintf (stdout, "         kB/min: %10.3f\n", kbps*60);
  fprintf (stdout, "           kB/h: %10.3f\n", kbps*3600);
  fprintf (stdout, "\n");

  return 0;
}
